# frozen_string_literal: true

require 'easy_type'

module Puppet_X
  module EnterpriseModules
    module Postgres
      # Daemonize the psql command
      class PsqlDaemon < PsqlCommand
        def initialize(options = {})
          super
          @catch_errors = Regexp.union(@catch_errors, /ORA-\d+:.*$/) if @failonsqlfail
          @daemon = EasyType::Daemon.run(identity)
          start_daemon if @daemon.nil?
        end

        # rubocop: disable Metrics/MethodLength
        def execute_sql_command(command, output_file, timeout = nil)
          timeout ||= @timeout
          ensure_connection
          begin
            Puppet.debug "Executing daemonized sql-command #{command}"
            @daemon.execute_command template('pg_config/pg_config/daemon_execute.sql.erb', binding)
            @daemon.sync(timeout) do |line|
              raise ThreadError if line.include?('ORA-03135')

              if line =~ @catch_errors
                @daemon.sync(timeout) # Clear buffer till end string
                fail "Error in execution of SQL command: #{command}.\nFound error #{line}"
              end
            end
          rescue ThreadError
            @retries = @retries.nil? ? 1 : @retries + 1
            fail 'maximum number of retries for daemon restart exhausted.' if @retries == 3

            Puppet.debug 'Detected disconnected daemon. restarting it...'
            @daemon.kill
            start_daemon
            retry
          end
          File.read(output_file)
        end
        # rubocop: enable Metrics/MethodLength

        private

        def start_daemon
          #
          # We don't need the sudo part of the returned command string, so
          # we remove it
          #
          @command = command_string.split('"')[1]
          Puppet.info "Starting the Psql daemon with os user #{@os_user} on db #{@db}"
          @daemon = EasyType::Daemon.new(identity, @command, @os_user)
          connect_to_postgres
        end

        #
        # This functions ensure's that the Psql proces is still running. Sometimes,
        # for example when postgres is beeing patched, the Psql process in the daemon stops.
        # This also means that Psql will not detect a ORA-03135 error because there is no Psql.
        # To fix this, for EVERY sql command we push to the daemon, we first send an empty
        # command to the daemon, with a short timeout (2 seconds). It just does a PROMPT.
        # If the Psql process is NOT running, it will stop the daemon and restart a new one.
        #
        def ensure_connection
          Puppet.debug 'Validating sql daemon'
          @daemon.execute_command template('pg_config/pg_config/validate_daemon.sql.erb', binding)
          loop do
            line = @daemon.send(:timed_readline, 2)
            break if line.include?('~~~~COMMAND SUCCESSFUL~~~~')
          end
        rescue StandardError
          Puppet.debug 'Detected daemon with Psql process. Restarting...'
          @daemon.kill
          start_daemon
        end

        def identity
          "postgres-#{@db}"
        end

        def connect_to_postgres
          Puppet.debug "Connecting to postgres db in daemon #{@db} with os_user #{@os_user}"
          @daemon.execute_command template('pg_config/pg_config/daemon_connect.sql.erb', binding)
        end
      end
    end
  end
end
